#include <stdio.h>
#include "allocate.h"
#include "multigrid.h"

static void exp_row(float *X, int M, float *IX);
static void exp_col(float **X, int N, float **IX, int col);
static void exp_lin_row(float *X, int M, float *IX);
static void exp_lin_col(float **X, int N, float **IX, int col);
static int interior(int i, int j, int N, int M);
static int within(int i, int j, int N, int M);
static int corner(int i, int j, int N, int M);


#define CP 0.125 
#define CN -0.125 
#define DCPCN  0.0

static int neigh[4][2] = {0, -1, 0, 1, -1, 0, 1, 0};
static float sg[4][2][2] = {CN, CP, CN, CP,
                            CP, CN, CP, CN,
                            CN, CN, CP, CP,
                            CP, CP, CN, CN};

static float H[4][4] = {0.25, 0.25, 0.25, 0.25,
					  0.25,-0.25, 0.25,-0.25,
					  0.25, 0.25,-0.25,-0.25,
					  0.25,-0.25,-0.25, 0.25};

void interp_quad(float **X1, int N, int M, float **X2)
{
    int i,j,ci,cj;
    float temp;

    for(ci=0;ci<N;ci++)
        for(cj=0;cj<M;cj++)
        {
            i = 2*ci;
            j = 2*cj;
            temp = X1[ci][cj];
            X2[i][j] = temp;
            X2[i][j+1] = temp;
            X2[i+1][j] = temp;
            X2[i+1][j+1] = temp;

        }
}


void decimate_quad(float **DX, int N, int M, float **X)
{
    int i,j,ci,cj;

    for(ci=0;ci<N;ci++)
        for(cj=0;cj<M;cj++)
        {
            i = 2*ci;
            j = 2*cj;
			DX[ci][cj] = ( X[i][j] + X[i][j+1] + X[i+1][j] + X[i+1][j+1])/4.0;
        }
}

void decimate_proj(float **DX, int N, int M, float **X)
{
    int i,j,ci,cj;

    for(ci=0;ci<N;ci++)
        for(cj=0;cj<M;cj++)
        {
            i = 2*ci;
            j = 2*cj;
            DX[ci][cj] = X[i][j] + X[i][j+1] + X[i+1][j] + X[i+1][j+1];
        }
}

void decimate_proj_var(float **DX, int N, int M, float **X, int dx, int dy)
{
    int i,j,ci,cj;

    for(ci=0;ci<N;ci++)
        for(cj=0;cj<M;cj++)
        {
            DX[ci][cj] = 0.0;
    		for(i=dx*ci;i<dx*ci+dx;i++)
    			for(j=dy*cj;j<dy*cj+dy;j++)
                  DX[ci][cj] += X[i][j];
        }
}



void decimate_mask(int **DX, int N, int M, int **X)
{
    int i,j,ci,cj;

    for(ci=0;ci<N;ci++)
        for(cj=0;cj<M;cj++)
        {
            i = 2*ci;
            j = 2*cj;
			if ( X[i][j] + X[i][j+1] + X[i+1][j] + X[i+1][j+1])
              DX[ci][cj] = 1;
			else
              DX[ci][cj] = 0;
        }
}



void interp_mmse(float **X1, float **Z, int N, int M, float **X2)
{
	int i,j,n,ni,nj,ii,jj,ci,cj;
	float coeff;


	for(i=0;i<4*N*M;i++)
		X2[0][i] = 0.0;

	for(i=0;i<N;i++)
		for(j=0;j<M;j++)
		{
			coeff = 1.0;
			if( within(i,j-1,N,M) && within(i,j+1,N,M))
				coeff += DCPCN;
			if( within(i-1,j,N,M) && within(i+1,j,N,M))
				coeff += DCPCN;
				
            for(ii=0;ii<2;ii++)
                for(jj=0;jj<2;jj++)
                {
					ci = 2*i+ii;
					cj = 2*j+jj;
					X2[ci][cj] += coeff*(X1[i][j]+Z[i][j]);
				}

			for(n=0;n<4;n++)
			{
				ni = i + 2*neigh[n][0];
				nj = j + 2*neigh[n][1];

				if( within(ni,nj,N,M) )
				{
					ni = i + neigh[n][0];
					nj = j + neigh[n][1];

            		for(ii=0;ii<2;ii++)
               			for(jj=0;jj<2;jj++)
                		{
							ci = 2*ni+ii;
							cj = 2*nj+jj;
							X2[ci][cj] += sg[n][ii][jj]*(X1[i][j]+Z[i][j]);
						}
				}
			}
		}
}


void interp(float **X, int N, int M, float **IX, int positivity)
{
	int i,j,n,ni,nj,ii,jj,ci,cj;
	float coeff;
	float **EX;

	EX = (float **) get_img(2*M, N, sizeof(float));

	if(M == 1 || M==2)
		for(i=0;i<N;i++)
			exp_lin_row(X[i],M,EX[i]);
	else
	{
		for(i=0;i<N;i++)
			exp_row(X[i],M,EX[i]);
		if(positivity)
			for(i=0;i<2*M*N;i++){
				if (EX[0][i] < 0.0)
					EX[0][i] = 0.0;
			}
	}	

	


	if(N == 1 || N==2)
		for(j=0;j<2*M;j++)
			exp_lin_col(EX,N,IX,j);
	else
	{
		for(j=0;j<2*M;j++)
			exp_col(EX,N,IX,j);
		if(positivity)
			for(i=0;i<4*M*N;i++){
				if (IX[0][i] < 0.0)
					IX[0][i] = 0.0;
			}
	}

}

static void exp_row(float *X, int M, float *IX)
{
	float lx,rx;
	int i,bi;
	static float c_odd[4]={-0.0703, 0.8672, 0.2266, -0.0234};
	static float c_even[4]={-0.0234, 0.2266, 0.8672, -0.0703};

	lx = X[2] - 3*X[1] + 3*X[0];
	rx = 3*X[M-1] - 3*X[M-2] + X[M-3];

	/*IX[0] = c_even[1]*lx + c_even[2]*X[0] + c_even[3]*X[1];*/
	IX[0] = X[0];
	IX[1] = c_odd[0]*lx + c_odd[1]*X[0] + c_odd[2]*X[1] + c_odd[3]*X[2];
	IX[2] = c_even[0]*lx + c_even[1]*X[0] + c_even[2]*X[1] + c_even[3]*X[2];
	/*IX[2*M-1] = c_odd[2]*rx + c_odd[1]*X[M-1] + c_odd[0]*X[M-2];*/
	IX[2*M-1] = X[M-1];
	IX[2*M-2] = c_even[3]*rx + c_even[2]*X[M-1] + c_even[1]*X[M-2] + c_even[0]*X[M-3];
	IX[2*M-3] = c_odd[3]*rx + c_odd[2]*X[M-1] + c_odd[1]*X[M-2] + c_odd[0]*X[M-3];


	for(i=3;i<2*M-3;i=i+2)
	{
		bi = i/2;
		IX[i] = c_odd[0]*X[bi-1] + c_odd[1]*X[bi] + c_odd[2]*X[bi+1] + c_odd[3]*X[bi+2];
	}

	for(i=4;i<2*M-3;i=i+2)
	{
		bi = i/2;
		IX[i] = c_even[0]*X[bi-2] + c_even[1]*X[bi-1] + c_even[2]*X[bi] + c_even[3]*X[bi+1];
	}
}

static void exp_col(float **X, int N, float **IX, int col)
{
	float lx,rx;
	int i,bi;
	static float c_odd[4]={-0.0703, 0.8672, 0.2266, -0.0234};
	static float c_even[4]={-0.0234, 0.2266, 0.8672, -0.0703};

	lx = X[2][col]- 3*X[1][col]+ 3*X[0][col];
	rx = 3*X[N-1][col]- 3*X[N-2][col]+ X[N-3][col];

	/*IX[0][col]= c_even[1]*lx + c_even[2]*X[0][col]+ c_even[3]*X[1][col];*/
	IX[0][col] = X[0][col];
	IX[1][col]= c_odd[0]*lx + c_odd[1]*X[0][col]+ c_odd[2]*X[1][col]+ c_odd[3]*X[2][col];
	IX[2][col]= c_even[0]*lx + c_even[1]*X[0][col]+ c_even[2]*X[1][col]+ c_even[3]*X[2][col];

	/*IX[2*N-1][col]= c_odd[2]*rx + c_odd[1]*X[N-1][col]+c_odd[0]*X[N-2][col];*/
	IX[2*N-1][col] = X[N-1][col];
	IX[2*N-2][col]= c_even[3]*rx + c_even[2]*X[N-1][col]+ c_even[1]*X[N-2][col]+ c_even[0]*X[N-3][col];
	IX[2*N-3][col]= c_odd[3]*rx + c_odd[2]*X[N-1][col]+ c_odd[1]*X[N-2][col]+ c_odd[0]*X[N-3][col];


	for(i=3;i<2*N-3;i=i+2)
	{
		bi = i/2;
		IX[i][col]= c_odd[0]*X[bi-1][col]+ c_odd[1]*X[bi][col]+ c_odd[2]*X[bi+1][col]+ c_odd[3]*X[bi+2][col];
	}

	for(i=4;i<2*N-3;i=i+2)
	{
		bi = i/2;
		IX[i][col]= c_even[0]*X[bi-2][col]+ c_even[1]*X[bi-1][col]+ c_even[2]*X[bi][col]+ c_even[3]*X[bi+1][col];
	}
}

static void exp_lin_row(float *X, int M, float *IX)
{

	if(M == 1)
	{
		IX[0] = X[0];
		IX[1] = X[0];
	}
	else
	{
		IX[0] = 0.75*X[0];
		IX[3] = 0.75*X[1];
		IX[1] = 0.75*X[0]+0.25*X[1];
		IX[2] = 0.75*X[1]+0.25*X[0];
	}
}

static void exp_lin_col(float **X, int N, float **IX, int col)
{

	if(N == 1)
	{
		IX[0][col] = X[0][col];
		IX[1][col] = X[0][col];
	}
	else
	{
		IX[0][col] = 0.75*X[0][col];
		IX[3][col] = 0.75*X[1][col];
		IX[1][col] = 0.75*X[0][col]+0.25*X[1][col];
		IX[2][col] = 0.75*X[1][col]+0.25*X[0][col];
	}
}

	



static int within(int i, int j, int N, int M)
{
	if ( i>=0 && i<N && j>=0 && j<M )
		return(1);
	else
		return(0);
}


static float B[2][2] = {0.5, 0.25, 0.75, 0.25}; 


/*
void interp_bilin(float **X1, float **Z, int N, int M, float **X2)
{
	int i,j,ii,jj,ci,new_ci,cj,cN,cM;

	cN = 2*N;
	cM = 2*M;

	for(i=0;i<cN*cM;i++)
		X2[0][i] = 0.0;

	for(i=0;i<N;i++)
		for(j=0;j<M;j++)
		{
            for(ii=0;ii<2;ii++)
                for(jj=0;jj<2;jj++)
                {
					ci = 2*i+ii;
					cj = 2*j+jj;

					if( interior(ci,cj,cN,cM))
					{
						X2[ci][cj] += B[0][0]*(X1[i][j]+Z[i][j]);
						X2[ci+odd(ci)][cj] += B[0][1]*(X1[i][j]+Z[i][j]);
						X2[ci][cj+odd(cj)] += B[0][1]*(X1[i][j]+Z[i][j]);
					}
					else if (corner(ci, cj, cN, cM))
						X2[ci][cj] += (X1[i][j]+Z[i][j]);
					else
					{
						X2[ci][cj] += B[1][0]*(X1[i][j]+Z[i][j]);
						new_ci = ci + odd(ci);
						if (new_ci >=0 && new_ci <cN)
							X2[new_ci][cj] += B[1][1]*(X1[i][j]+Z[i][j]);
						else
							X2[ci][cj+odd(cj)] += B[1][1]*(X1[i][j]+Z[i][j]);
					}


				}
		}
}

*/

void interp_bilin(float **X1, int N, int M, float **X2)
{
    int i,j,ii,jj,ci,new_ci,cj,cN,cM;

    cN = 2*N;
    cM = 2*M;

    for(i=0;i<cN*cM;i++)
        X2[0][i] = 0.0;

    for(i=0;i<N;i++)
        for(j=0;j<M;j++)
        {
            for(ii=0;ii<2;ii++)
                for(jj=0;jj<2;jj++)
                {
                    ci = 2*i+ii;
                    cj = 2*j+jj;

                    if( interior(ci,cj,cN,cM))
                    {
                        X2[ci][cj] += B[0][0]*X1[i][j];
                        X2[ci+odd(ci)][cj] += B[0][1]*X1[i][j];
                        X2[ci][cj+odd(cj)] += B[0][1]*X1[i][j];
                    }
                    else if (corner(ci, cj, cN, cM))
                        X2[ci][cj] += X1[i][j];
                    else
                    {
                        X2[ci][cj] += B[1][0]*X1[i][j];
                        new_ci = ci + odd(ci);
                        if (new_ci >=0 && new_ci <cN)
                            X2[new_ci][cj] += B[1][1]*X1[i][j];
                        else
                            X2[ci][cj+odd(cj)] += B[1][1]*X1[i][j];
                    }


                }
        }
}




void propagate_R(float **Z, float **R1, float **R2, int N, int M)
{
	int i,j,ci,cj,ii,jj;
	float max;

	for(i=0;i<N;i++)
		for(j=0;j<M;j++)
		{
			max = 0.0;
			for(ii=0;ii<2;ii++)
				for(jj=0;jj<2;jj++)
				{
					ci = 2*i+ii;
					cj = 2*j+jj;

					if (max < R1[ci][cj] - Z[ci][cj])
						max = R1[ci][cj] - Z[ci][cj];
				}

			R2[i][j] = max;
		}
}

			
static int odd(int i)
{
    if( (i/2)*2 == i)
        return(-1);
    else
        return(1);
}




static int interior(int i, int j, int N, int M)
{
	if ( i>0 && i<N-1 && j>0 && j<M-1 )
		return(1);
	else
		return(0);
}

static int corner(int i, int j, int N, int M)
{
	if( i==0)
	{	if( j==0)
			return(1);
		if( j==M-1)
			return(1);
	}
	else if( i==N-1)
	{
		if( j==0)
			return(1);
		if( j==M-1)
			return(1);
	}	

	return(0);
}
			

#undef CP
#undef CN
#undef DCPCN
